Loosen the requirements for atomic and legacy commit so that, in cases
where pixel_source != FB, the commit can still go through.

This includes adding framebuffer NULL checks in other areas to account for
FB being NULL when non-FB pixel sources are enabled.

To disable a plane, the pixel_source must be NONE or the FB must be NULL
if pixel_source == FB.

Signed-off-by: Jessica Zhang <quic_jessz...@quicinc.com>
---
 drivers/gpu/drm/drm_atomic.c        | 21 ++++++++++----------
 drivers/gpu/drm/drm_atomic_helper.c | 39 +++++++++++++++++++++----------------
 include/drm/drm_atomic_helper.h     |  4 ++--
 include/drm/drm_plane.h             | 29 +++++++++++++++++++++++++++
 4 files changed, 64 insertions(+), 29 deletions(-)

diff --git a/drivers/gpu/drm/drm_atomic_helper.c 
b/drivers/gpu/drm/drm_atomic_helper.c
index 292e38eb6218..27e8c8dd9324 100644
--- a/drivers/gpu/drm/drm_atomic_helper.c
+++ b/drivers/gpu/drm/drm_atomic_helper.c
@@ -852,7 +852,6 @@ int drm_atomic_helper_check_plane_state(struct 
drm_plane_state *plane_state,
                                        bool can_position,
                                        bool can_update_disabled)
 {
-       struct drm_framebuffer *fb = plane_state->fb;
        struct drm_rect *src = &plane_state->src;
        struct drm_rect *dst = &plane_state->dst;
        unsigned int rotation = plane_state->rotation;
@@ -864,7 +863,7 @@ int drm_atomic_helper_check_plane_state(struct 
drm_plane_state *plane_state,
        *src = drm_plane_state_src(plane_state);
        *dst = drm_plane_state_dest(plane_state);
 
-       if (!fb) {
+       if (!drm_plane_has_visible_data(plane_state)) {
                plane_state->visible = false;
                return 0;
        }
@@ -881,25 +880,31 @@ int drm_atomic_helper_check_plane_state(struct 
drm_plane_state *plane_state,
                return -EINVAL;
        }
 
-       drm_rect_rotate(src, fb->width << 16, fb->height << 16, rotation);
+       /* Check that selected pixel source is valid */
+       if (plane_state->pixel_source == DRM_PLANE_PIXEL_SOURCE_FB && 
plane_state->fb) {
+               struct drm_framebuffer *fb = plane_state->fb;
+               drm_rect_rotate(src, fb->width << 16, fb->height << 16, 
rotation);
 
-       /* Check scaling */
-       hscale = drm_rect_calc_hscale(src, dst, min_scale, max_scale);
-       vscale = drm_rect_calc_vscale(src, dst, min_scale, max_scale);
-       if (hscale < 0 || vscale < 0) {
-               drm_dbg_kms(plane_state->plane->dev,
-                           "Invalid scaling of plane\n");
-               drm_rect_debug_print("src: ", &plane_state->src, true);
-               drm_rect_debug_print("dst: ", &plane_state->dst, false);
-               return -ERANGE;
-       }
+               /* Check scaling */
+               hscale = drm_rect_calc_hscale(src, dst, min_scale, max_scale);
+               vscale = drm_rect_calc_vscale(src, dst, min_scale, max_scale);
 
-       if (crtc_state->enable)
-               drm_mode_get_hv_timing(&crtc_state->mode, &clip.x2, &clip.y2);
+               if (hscale < 0 || vscale < 0) {
+                       drm_dbg_kms(plane_state->plane->dev,
+                                       "Invalid scaling of plane\n");
+                       drm_rect_debug_print("src: ", &plane_state->src, true);
+                       drm_rect_debug_print("dst: ", &plane_state->dst, false);
+                       return -ERANGE;
+               }
 
-       plane_state->visible = drm_rect_clip_scaled(src, dst, &clip);
+               if (crtc_state->enable)
+                       drm_mode_get_hv_timing(&crtc_state->mode, &clip.x2, 
&clip.y2);
 
-       drm_rect_rotate_inv(src, fb->width << 16, fb->height << 16, rotation);
+               plane_state->visible = drm_rect_clip_scaled(src, dst, &clip);
+               drm_rect_rotate_inv(src, fb->width << 16, fb->height << 16, 
rotation);
+       } else if (drm_plane_solid_fill_enabled(plane_state)) {
+               plane_state->visible = true;
+       }
 
        if (!plane_state->visible)
                /*
diff --git a/include/drm/drm_atomic_helper.h b/include/drm/drm_atomic_helper.h
index 536a0b0091c3..6d97f38ac1f6 100644
--- a/include/drm/drm_atomic_helper.h
+++ b/include/drm/drm_atomic_helper.h
@@ -256,8 +256,8 @@ drm_atomic_plane_disabling(struct drm_plane_state 
*old_plane_state,
         * Anything else should be considered a bug in the atomic core, so we
         * gently warn about it.
         */
-       WARN_ON((new_plane_state->crtc == NULL && new_plane_state->fb != NULL) 
||
-               (new_plane_state->crtc != NULL && new_plane_state->fb == NULL));
+       WARN_ON((new_plane_state->crtc == NULL && 
drm_plane_has_visible_data(new_plane_state)) ||
+               (new_plane_state->crtc != NULL && 
!drm_plane_has_visible_data(new_plane_state)));
 
        return old_plane_state->crtc && !new_plane_state->crtc;
 }
diff --git a/include/drm/drm_plane.h b/include/drm/drm_plane.h
index 6171fb1a0b47..56b7a0acce74 100644
--- a/include/drm/drm_plane.h
+++ b/include/drm/drm_plane.h
@@ -992,6 +992,35 @@ static inline struct drm_plane *drm_plane_find(struct 
drm_device *dev,
 #define drm_for_each_plane(plane, dev) \
        list_for_each_entry(plane, &(dev)->mode_config.plane_list, head)
 
+/**
+ * drm_plane_solid_fill_enabled - Check if solid fill is enabled on plane
+ * @state: plane state
+ *
+ * Returns:
+ * Whether the plane has been assigned a solid_fill_blob
+ */
+static inline bool drm_plane_solid_fill_enabled(struct drm_plane_state *state)
+{
+       if (!state)
+               return false;
+       return state->pixel_source == DRM_PLANE_PIXEL_SOURCE_SOLID_FILL && 
state->solid_fill_blob;
+}
+
+static inline bool drm_plane_has_visible_data(const struct drm_plane_state 
*state)
+{
+       switch (state->pixel_source) {
+       case DRM_PLANE_PIXEL_SOURCE_NONE:
+               return false;
+       case DRM_PLANE_PIXEL_SOURCE_SOLID_FILL:
+               return state->solid_fill_blob != NULL;
+       case DRM_PLANE_PIXEL_SOURCE_FB:
+       default:
+               WARN_ON(state->pixel_source != DRM_PLANE_PIXEL_SOURCE_FB);
+       }
+
+       return state->fb != NULL;
+}
+
 bool drm_any_plane_has_format(struct drm_device *dev,
                              u32 format, u64 modifier);
 

diff --git a/drivers/gpu/drm/drm_atomic.c b/drivers/gpu/drm/drm_atomic.c
index 48a2df4e3d27..66e49c06aced 100644
--- a/drivers/gpu/drm/drm_atomic.c
+++ b/drivers/gpu/drm/drm_atomic.c
@@ -674,17 +674,16 @@ static int drm_atomic_plane_check(const struct 
drm_plane_state *old_plane_state,
 {
        struct drm_plane *plane = new_plane_state->plane;
        struct drm_crtc *crtc = new_plane_state->crtc;
-       const struct drm_framebuffer *fb = new_plane_state->fb;
        int ret;
 
-       /* either *both* CRTC and FB must be set, or neither */
-       if (crtc && !fb) {
-               drm_dbg_atomic(plane->dev, "[PLANE:%d:%s] CRTC set but no FB\n",
+       /* either *both* CRTC and pixel source must be set, or neither */
+       if (crtc && !drm_plane_has_visible_data(new_plane_state)) {
+               drm_dbg_atomic(plane->dev, "[PLANE:%d:%s] CRTC set but no 
visible data\n",
                               plane->base.id, plane->name);
                return -EINVAL;
-       } else if (fb && !crtc) {
-               drm_dbg_atomic(plane->dev, "[PLANE:%d:%s] FB set but no CRTC\n",
-                              plane->base.id, plane->name);
+       } else if (drm_plane_has_visible_data(new_plane_state) && !crtc) {
+               drm_dbg_atomic(plane->dev, "[PLANE:%d:%s] Source %d has visible 
data but no CRTC\n",
+                              plane->base.id, plane->name, 
new_plane_state->pixel_source);
                return -EINVAL;
        }
 
@@ -715,9 +714,11 @@ static int drm_atomic_plane_check(const struct 
drm_plane_state *old_plane_state,
        }
 
 
-       ret = drm_atomic_plane_check_fb(new_plane_state);
-       if (ret)
-               return ret;
+       if (new_plane_state->pixel_source == DRM_PLANE_PIXEL_SOURCE_FB && 
new_plane_state->fb) {
+               ret = drm_atomic_plane_check_fb(new_plane_state);
+               if (ret)
+                       return ret;
+       }
 
        if (plane_switching_crtc(old_plane_state, new_plane_state)) {
                drm_dbg_atomic(plane->dev,
-- 
2.42.0

Reply via email to